import abc
import logging
import os
import queue
import threading
import time
import uuid
import weakref

TIME_SLEEP = 0.001
TIMEOUT = 1.0
TIMEOUT_WORKER = 20

logger = logging.getLogger()


# logger.setLevel("INFO")


class FutureCache(dict):
    # Dict for weakref only
    pass


class Future:
    def __init__(self, task_id, task_size: int, future_cache_ref):
        self._id = task_id
        self._size = task_size
        self._future_cache_ref = future_cache_ref
        self._outputs = []
        self._finish_event = threading.Event()

    def done(self):
        return self._finish_event.is_set()

    def result(self, timeout=None):
        if self._size == 0:
            self._finish_event.set()
            return []
        finished = self._finish_event.wait(timeout)

        if not finished:
            raise TimeoutError("Task: %d Timeout" % self._id)

        # remove from future_cache
        future_cache = self._future_cache_ref()
        if future_cache is not None:
            del future_cache[self._id]

        # [(request_id, output), ...] sorted by request_id
        self._outputs.sort(key=lambda i: i[0])
        # restore batch result from outputs
        batch_result = [i[1] for i in self._outputs]

        return batch_result

    def append_result(self, it_id, it_output):
        self._outputs.append((it_id, it_output))
        if len(self._outputs) >= self._size:
            self._finish_event.set()


class BaseWorker(abc.ABC):
    __slots__ = ('_request_queue', '_response_queue')

    def __init__(self, func_predict: callable, batch_size: int, max_latency: float, destroy_event=None, *args, **kwargs):
        # assert callable(predict_function)
        self.pid = os.getpid()
        self.func_predict = func_predict
        self.batch_size = batch_size
        self.max_latency = max_latency
        self.destroy_event = destroy_event  # kwargs.get("destroy_event", None)

    def _recv_request(self, timeout: float = TIMEOUT):
        try:
            item = self._request_queue.get(timeout=timeout)
        except queue.Empty:
            raise TimeoutError
        else:
            return item

    def _send_response(self, client_id, task_id, request_id, model_output):
        self._response_queue.put((task_id, request_id, model_output))

    def run_forever(self, *args, **kwargs):
        self.pid = os.getpid()  # overwrite the pid
        logger.info("[gpu worker %d] %s start working" % (self.pid, self))

        while True:
            handled = self._run_once()
            if self.destroy_event and self.destroy_event.is_set():
                break
            if not handled:
                # sleep if no data handled last time
                time.sleep(TIME_SLEEP)
        logger.info("[gpu worker %d] %s shutdown" % (self.pid, self))

    def model_predict(self, batch_input):
        batch_result = self.func_predict(batch_input)
        assert len(batch_input) == len(batch_result), "input batch size {} and output batch size {} must be equal.".format(len(batch_input), len(batch_result))
        return batch_result

    def _run_once(self):
        batch = []
        start_time = time.time()
        for i in range(self.batch_size):
            try:
                item = self._recv_request(timeout=self.max_latency)
            except TimeoutError:
                # each item timeout exceed the max latency
                break
            else:
                batch.append(item)
            if (time.time() - start_time) > self.max_latency:
                # total batch time exceeds the max latency
                break
        if not batch:
            return 0

        model_inputs = [i[3] for i in batch]
        model_outputs = self.model_predict(model_inputs)

        # send results to response
        for i, item in enumerate(batch):
            client_id, task_id, request_id, _ = item
            self._send_response(client_id, task_id, request_id, model_outputs[i])

        batch_size = len(batch)
        logger.info("[gpu worker %d] run_once batch_size: %d start_at: %s spend: %s" % (
            self.pid, batch_size, start_time, time.time() - start_time))
        return batch_size


class BaseStreamer(abc.ABC):
    __slots__ = ('_input_queue', '_output_queue')

    def __init__(self, worker_timeout: float = TIMEOUT_WORKER, *args, **kwargs):
        self._client_id = str(uuid.uuid4())
        self._task_id = 0
        self._future_cache = FutureCache()  # {task_id: future}
        self._worker_timeout = worker_timeout  # kwargs.get("worker_timeout", TIMEOUT_WORKER)

        self.back_thread = threading.Thread(target=self._loop_collect_result, name="thread_collect_result")
        self.back_thread.daemon = True
        self.lock = threading.Lock()

    def _send_request(self, task_id, request_id, model_input):
        self._input_queue.put((0, task_id, request_id, model_input))

    def _recv_response(self, timeout=TIMEOUT):
        try:
            message = self._output_queue.get(timeout=timeout)
        except queue.Empty:
            message = None
        return message

    @abc.abstractmethod
    def destroy_workers(self):
        raise NotImplementedError

    def _delay_setup(self):
        self.back_thread.start()

    def _input(self, batch: list) -> int:
        """input a batch, distribute each item to mq, return task_id"""
        # task id in one client
        self.lock.acquire()
        task_id = self._task_id
        self._task_id += 1
        self.lock.release()
        # request id in one task
        request_id = 0

        future = Future(task_id, len(batch), weakref.ref(self._future_cache))
        self._future_cache[task_id] = future

        for model_input in batch:
            self._send_request(task_id, request_id, model_input)
            request_id += 1

        return task_id

    def _loop_collect_result(self):
        logger.info("start _loop_collect_result")
        while True:
            message = self._recv_response(timeout=TIMEOUT)
            if message:
                (task_id, request_id, item) = message
                future = self._future_cache[task_id]
                future.append_result(request_id, item)
            else:
                # todo
                time.sleep(TIME_SLEEP)

    def _output(self, task_id: int) -> list:
        future = self._future_cache[task_id]
        batch_result = future.result(self._worker_timeout)
        return batch_result

    def submit(self, batch):
        task_id = self._input(batch)
        future = self._future_cache[task_id]
        return future

    def predict(self, batch):
        task_id = self._input(batch)
        ret = self._output(task_id)
        assert len(batch) == len(ret), "input batch size {} and output batch size {} must be equal.".format(len(batch), len(ret))
        return ret



import pyhocon

from .base import BaseVault
from ...encrypt.aes import AesEncryptor


def _is_empty_str(s):
    return s is None or isinstance(s, pyhocon.config_tree.NoneValue) or s == "None" or s == ""


class AesVault(AesEncryptor, BaseVault):
    def __init__(self, key: str = None):
        super().__init__(key)

    def decrypt_password(self, pwd):
        if _is_empty_str(pwd):
            return None
        return self.decrypt(pwd)


def main():
    vault = AesVault()
    pwd = vault.get_password(None, url_quote=True)
    # print(pwd)
    return pwd



__all__ = ('get_gpu_info',)

# ! `pip install pynvml`, reference: https://github.com/gpuopenanalytics/pynvml

from collections import namedtuple

from ..logger import LOG

try:
    import pynvml as nvml
    from pynvml.smi import nvidia_smi

    LOG.debug('Using pynvml == %s' % nvml.__version__)
except ImportError:
    LOG.error('Package `pynvml` NOT installed! Cannot get GPU info.')
    nvml = nvidia_smi = None

Device = namedtuple('Device', field_names='index,name,arch')
DeviceStatus = namedtuple('DeviceStatus', field_names='mem_total,mem_free,mem_used,gpu_rate,mem_rate')


class NvInfo:
    def __init__(self):
        global nvml
        if nvml is None:
            return
        try:
            nvml.nvmlInit()
            LOG.debug('NVML loaded and initialized successfully.')
        except Exception as e:
            LOG.error('Fail to initialize NVML!')
            nvml = None

    def __del__(self):
        try:
            if nvml is not None:
                nvml.nvmlShutdown()
        except Exception as e:
            LOG.error('Exception removing NvInfo: %s' % e)

    @staticmethod
    def get_driver_version() -> str:
        if nvml is not None:
            try:
                ver: bytes = nvml.nvmlSystemGetDriverVersion()
                LOG.debug('GPU driver version %s' % str(ver))
                return ver.decode(encoding='UTF-8')
            except Exception as e:
                LOG.info('NVML library error: %s' % str(e))
        return 'Unknown'

    @staticmethod
    def get_device_count() -> int:
        try:
            return nvml.nvmlDeviceGetCount()
        except Exception as e:
            LOG.info('NVML library error: %s' % str(e))
            return 0

    def get_device_list(self) -> list:
        list_device = []
        for i in range(self.get_device_count()):
            handler = self._get_device_handler(index_device=i)
            name, arch = None, None

            try:
                name = nvml.nvmlDeviceGetName(handler).decode(encoding='UTF-8')
            except Exception as e:
                LOG.info('Failed to get device name!')
                LOG.info(str(e))

            try:
                arch = nvml.nvmlDeviceGetArchitecture(handler)
            except Exception as e:
                LOG.info('Failed to get device architecture!')
                LOG.info(str(e))

            device = Device(index=i, name=name, arch=arch)
            LOG.debug('Found device {index} info: name={name}; arch={arch}'.format(**device._asdict()))
            list_device.append(device)

        return list_device

    def _get_device_handler(self, index_device=0):
        return nvml.nvmlDeviceGetHandleByIndex(index_device)

    def get_device_status(self, index_device=0) -> DeviceStatus:
        handler = self._get_device_handler(index_device=index_device)
        mem_info = nvml.nvmlDeviceGetMemoryInfo(handler)
        util_info = nvml.nvmlDeviceGetUtilizationRates(handler)
        return DeviceStatus(
            mem_total=mem_info.total,
            mem_free=mem_info.free,
            mem_used=mem_info.used,
            gpu_rate=util_info.gpu,
            mem_rate=util_info.memory
        )

    @staticmethod
    def get_smi():
        if nvidia_smi is None:
            return
        try:
            return nvidia_smi.getInstance()
        except Exception as e:
            LOG.warn('Failed to get smi: %s' % str(e))
            return


def get_gpu_info(*args, **kwargs) -> dict:
    nv_info = NvInfo()
    list_device = []
    list_devices = nv_info.get_device_list()
    for i, d in enumerate(list_devices):
        status = nv_info.get_device_status(index_device=i)
        item = {
            'device': d._asdict(),
            'status': status._asdict()
        }
        list_device.append(item)

    ret = {
        "driver_version": nv_info.get_driver_version(),
        "devices": list_device,
    }

    smi = nv_info.get_smi()
    if smi is not None:
        if len(args) == 0:
            args = 'name;vbios_version;inforom.oem;compute-apps'.split(';')
        for k in args:
            ret[k] = smi.DeviceQuery(k)

    return {"gpu_info": ret}


def main(*args, **kwargs):
    info_gpus = get_gpu_info()
    import json
    print(json.dumps(info_gpus, ensure_ascii=False, indent=2))


if __name__ == '__main__':
    main()



import multiprocessing
import os

from .base import BaseStreamer, BaseWorker, TIMEOUT_WORKER, logger
from .managed_model import ManagedModel


class MultiProcessWorker(BaseWorker):
    def __init__(self, func_predict: callable, batch_size: int, max_latency: float,
                 request_queue, response_queue, model_init_args, model_init_kwargs: dict, *args, **kwargs):
        super().__init__(func_predict, batch_size, max_latency, *args, **kwargs)
        self._request_queue = request_queue
        self._response_queue = response_queue
        self._model_init_args = model_init_args or []
        self._model_init_kwargs = model_init_kwargs or {}

    def run_forever(self, gpu_id=None, ready_event=None, destroy_event=None):
        # if it is a managed model, lazy init model after forked & set CUDA_VISIBLE_DEVICES
        if isinstance(self.func_predict, type) and issubclass(self.func_predict, ManagedModel):
            model_class = self.func_predict
            logger.info("[gpu worker %d] init model on gpu:%s" % (os.getpid(), gpu_id))
            self._model = model_class(gpu_id)
            self._model.init_model(*self._model_init_args, **self._model_init_kwargs)
            logger.info("[gpu worker %d] init model on gpu:%s" % (os.getpid(), gpu_id))
            self._predict = self._model.predict
            if ready_event:
                ready_event.set()  # tell father process that init is finished
        if destroy_event is not None:
            self.destroy_event = destroy_event
        super().run_forever()


class MultiProcessStreamer(BaseStreamer):
    def __init__(self, func_predict: callable, batch_size: int, max_latency: float = 0.1, worker_num: int = 1,
                 cuda_devices=None, model_init_args=None, model_init_kwargs: dict = None, wait_for_worker_ready: bool = False,
                 mp_start_method='spawn', worker_timeout: float = TIMEOUT_WORKER):
        super().__init__(worker_timeout=worker_timeout)

        self.multiprocess = multiprocessing.get_context(mp_start_method)
        self._input_queue = self.multiprocess.Queue()
        self._output_queue = self.multiprocess.Queue()

        self.worker_num = worker_num
        self.cuda_devices = cuda_devices

        self._worker = MultiProcessWorker(
            func_predict, batch_size, max_latency, self._input_queue, self._output_queue,
            model_init_args, model_init_kwargs
        )
        self._worker_processes = []
        self._worker_ready_events = []
        self._worker_destroy_events = []
        self._setup_gpu_worker()
        if wait_for_worker_ready:
            self._wait_for_worker_ready()
        self._delay_setup()

    def _setup_gpu_worker(self):
        for i in range(self.worker_num):
            ready_event = self.multiprocess.Event()
            destroy_event = self.multiprocess.Event()
            if self.cuda_devices is not None:
                gpu_id = self.cuda_devices[i % len(self.cuda_devices)]
                args = (gpu_id, ready_event, destroy_event)
            else:
                args = (None, ready_event, destroy_event)
            p = self.multiprocess.Process(target=self._worker.run_forever, args=args, name="stream_worker", daemon=True)
            p.start()
            self._worker_processes.append(p)
            self._worker_ready_events.append(ready_event)
            self._worker_destroy_events.append(destroy_event)

    def _wait_for_worker_ready(self, timeout=None):
        if timeout is None:
            timeout = self._worker_timeout
        # wait for all workers finishing init
        for i, e in enumerate(self._worker_ready_events):
            # todo: select all events with timeout
            is_ready = e.wait(timeout)
            logger.info("gpu worker:%d ready state: %s" % (i, is_ready))

    def destroy_workers(self):
        for e in self._worker_destroy_events:
            e.set()
        for p in self._worker_processes:
            p.join(timeout=self._worker_timeout)
            if p.is_alive():
                raise TimeoutError("worker_process destroy timeout")
        logger.info("workers destroyed")



import multiprocessing
import os
import pickle
import queue
import threading
import time

from redis import Redis

from .base import BaseStreamer, BaseWorker, TIMEOUT, TIME_SLEEP, logger


class RedisWorker(BaseWorker):
    def __init__(self, func_predict, batch_size: int, max_latency=0.1,
                 redis_broker="localhost:6379", prefix='',
                 model_init_args=None, model_init_kwargs=None, *args, **kwargs):
        # assert issubclass(model_class, ManagedModel)
        super().__init__(func_predict, batch_size, max_latency, *args, **kwargs)
        self._request_queue = queue.Queue()
        # redis worker does not need a response queue

        self.prefix = prefix
        self._model_init_args = model_init_args or []
        self._model_init_kwargs = model_init_kwargs or {}
        self._redis_broker = redis_broker
        self._redis = _RedisServer(0, self._redis_broker, self.prefix)

        self.back_thread = threading.Thread(target=self._loop_recv_request, name="thread_recv_request")
        self.back_thread.daemon = True
        self.back_thread.start()

    def _send_response(self, client_id, task_id, request_id, model_output):
        # override the parent method
        self._redis.send_response(client_id, task_id, request_id, model_output)

    def run_forever(self, gpu_id=None):
        logger.info("[gpu worker %d] init model on gpu:%s" % (os.getpid(), gpu_id))
        model_class = self.func_predict
        self._model = model_class(gpu_id)
        self._model.init_model(*self._model_init_args, **self._model_init_kwargs)
        self._predict = self._model.predict

        super().run_forever()

    def _loop_recv_request(self):
        logger.info("[gpu worker %d] start loop_recv_request" % (os.getpid()))
        while True:
            message = self._redis.recv_request(timeout=TIMEOUT)
            if message:
                (client_id, task_id, request_id, request_item) = pickle.loads(message)
                self._request_queue.put((client_id, task_id, request_id, request_item))
            else:
                # sleep if recv timeout
                time.sleep(TIME_SLEEP)


class RedisStreamer(BaseStreamer):
    """
    1. input batch as a task
    2. distribute every single item in batch to redis
    3. backend loop collecting results
    3. output batch result for a task when every single item is returned
    """

    def __init__(self, redis_broker="localhost:6379", prefix=''):
        super().__init__()

        # redis streamer does not need input_queue/output_queue

        self.prefix = prefix
        self._redis_broker = redis_broker
        self._redis = _RedisClient(self._client_id, self._redis_broker, self.prefix)
        self._delay_setup()

    def _send_request(self, task_id, request_id, model_input):
        self._redis.send_request(task_id, request_id, model_input)

    def _recv_response(self, timeout=TIMEOUT):
        return self._redis.recv_response(timeout)

    def destroy_workers(self):
        pass


def _setup_redis_worker_and_runforever(
        model_class, batch_size, max_latency, gpu_id, redis_broker, prefix='', model_init_args=None, model_init_kwargs=None
):
    redis_worker = RedisWorker(
        model_class, batch_size, max_latency, redis_broker=redis_broker, prefix=prefix,
        model_init_args=model_init_args, model_init_kwargs=model_init_kwargs
    )
    redis_worker.run_forever(gpu_id)


def run_redis_workers_forever(
        model_class, batch_size, max_latency=0.1,
        worker_num=1, cuda_devices=None, redis_broker="localhost:6379",
        prefix='', mp_start_method='spawn', model_init_args=None, model_init_kwargs=None
):
    procs = []
    mp = multiprocessing.get_context(mp_start_method)
    for i in range(worker_num):
        if cuda_devices is not None:
            gpu_id = cuda_devices[i % len(cuda_devices)]
        else:
            gpu_id = None
        args = (model_class, batch_size, max_latency, gpu_id, redis_broker, prefix, model_init_args, model_init_kwargs)
        p = mp.Process(target=_setup_redis_worker_and_runforever, args=args, name="stream_worker", daemon=True)
        p.start()
        procs.append(p)

    for p in procs:
        p.join()


class _RedisAgent:
    def __init__(self, redis_id, redis_broker='localhost:6379', prefix=''):
        self._redis_id = redis_id
        self._redis_host = redis_broker.split(":")[0]
        self._redis_port = int(redis_broker.split(":")[1])
        self._redis_request_queue_name = "request_queue" + prefix
        self._redis_response_pb_prefix = "response_pb_" + prefix
        self._redis = Redis(host=self._redis_host, port=self._redis_port)
        self._response_pb = self._redis.pubsub(ignore_subscribe_messages=True)
        self._setup()

    def _setup(self):
        raise NotImplementedError

    def _response_pb_name(self, redis_id):
        return self._redis_response_pb_prefix + redis_id


class _RedisClient(_RedisAgent):
    def _setup(self):
        self._response_pb.subscribe(self._response_pb_name(self._redis_id))

    def send_request(self, task_id, request_id, model_input):
        message = (self._redis_id, task_id, request_id, model_input)
        self._redis.lpush(self._redis_request_queue_name, pickle.dumps(message))

    def recv_response(self, timeout):
        message = self._response_pb.get_message(timeout=timeout)
        if message:
            return pickle.loads(message["data"])


class _RedisServer(_RedisAgent):
    def _setup(self):
        self._response_pb.psubscribe(self._redis_response_pb_prefix + "*")  # server subscribe all pubsub

    def recv_request(self, timeout):
        message = self._redis.blpop(self._redis_request_queue_name, timeout=timeout)
        if message:  # (queue_name, data)
            return message[1]

    def send_response(self, client_id, task_id, request_id, model_output):
        message = (task_id, request_id, model_output)
        channel_name = self._response_pb_name(client_id)
        self._redis.publish(channel_name, pickle.dumps(message))



import abc
import json
import logging
from datetime import datetime
from typing import Optional, Awaitable

from tornado import web

from ...encrypt import jwt
from ...logger import LOG
from ...settings import SETTINGS

_RESP_BAD_REQUEST = {'code': '5101', 'message': ['Bad request: fail to parse body as JSON object!']}


class APIHandler(web.RequestHandler):
    LOG = LOG

    def __init__(self, *args, **kwargs):
        self.api_args: Optional[tuple] = None
        self.api_kwargs: Optional[dict] = None
        super().__init__(*args, **kwargs)

    @abc.abstractmethod
    def response(self, *args, **kwargs) -> dict:
        raise NotImplementedError()

    @property
    def request_id(self):
        if 'Request-ID' not in self.request.headers:
            self.request.headers['Request-ID'] = datetime.now().strftime('%Y%m%d-%H%M%S-%f')
        return self.request.headers.get('Request-ID')

    def set_default_headers(self) -> None:
        self.set_header('Content-Type', 'application/json; charset=utf-8')

    def data_received(self, chunk: bytes) -> Optional[Awaitable[None]]:
        pass

    def on_finish(self) -> None:
        func_callback = getattr(self, 'callback', None)
        if not callable(func_callback):
            return

        return func_callback(*self.api_args, **self.api_kwargs)

    async def prepare(self, ) -> Optional[Awaitable[None]]:
        access_token = self.request.headers.get('Access-Token')
        if access_token is None:
            return self.finish({
                'msg': 'Invalid Access-Token in request header!'
            })
        else:
            secret_key = SETTINGS.config['APP_SECRET_KEY']  # 'HCTECH-ASKBOB-REC:10062462'
            # options = None
            # TODO: if not validate expiration
            options = {"verify_exp": False}
            access_token = jwt.decode(secret_key, access_token, options=options)
            if not isinstance(access_token, dict):
                LOG.error('Invalid Access-Token found in request for [%s]: %s' % (
                    str(self.request.full_url()), access_token
                ))
                return self.finish({
                    'msg': access_token
                })

        self.set_header('Request-ID', self.request_id)

    async def get(self, *args, **kwargs):
        query = {k: v[0].decode('utf-8') for k, v in self.request.arguments.items()}
        kwargs.update(query)
        try:
            self.LOG.debug('GET Request [%s]: %s' % (self.request_id, kwargs))
            self.api_args, self.api_kwargs = args or (), kwargs or {}
            resp = self.response(*self.api_args, **self.api_kwargs)  # this call may throw TypeError when argument missing
        except Exception as e:
            self.LOG.error(e, exc_info=True)
            self.LOG.info('GET Request [%s]: %s' % (self.request_id, kwargs))
            return self.finish({'status': 'error', 'message': [str(e)]})

        if isinstance(resp, (dict, list)):
            resp = json.dumps(resp, ensure_ascii=False, default=str, separators=(',', ':'))
        elif isinstance(resp, str):
            pass
        return self.finish(resp)

    async def post(self, *args, **kwargs):
        try:
            body = self.request.body.decode('utf-8')
            data = json.loads(body)
            kwargs.update(data)
        except json.decoder.JSONDecodeError:  # invalid request body, cannot be parsed as JSON
            return self.finish(_RESP_BAD_REQUEST)

        try:
            if self.LOG.level == logging.DEBUG:
                s_kwargs = json.dumps(kwargs, ensure_ascii=False)
                self.LOG.debug('POST Request [%s]: %s' % (self.request_id, s_kwargs[:1000]))
            self.api_args, self.api_kwargs = args or (), kwargs or {}
            resp = self.response(*self.api_args, **self.api_kwargs)  # this call may throw TypeError when argument missing
        except Exception as e:
            self.LOG.error(e, exc_info=True)
            self.LOG.info('POST Request [%s]: %s' % (self.request_id, data))
            return self.finish({'status': 'error', 'message': [str(e)]})

        if isinstance(resp, (dict, list)):
            resp = json.dumps(resp, ensure_ascii=False, default=str, separators=(',', ':'))
        elif isinstance(resp, str):
            pass
        return self.finish(resp)



__all__ = ('RsaEncryptor',)

import base64
from functools import lru_cache
from typing import Tuple, Optional, Union

from Crypto.Cipher import PKCS1_OAEP, PKCS1_v1_5
from Crypto.Hash import SHA1, SHA256
from Crypto.PublicKey import RSA
from Crypto.Signature import pss

t_cipher_module = Union[PKCS1_v1_5.PKCS115_Cipher, PKCS1_OAEP.PKCS1OAEP_Cipher]

_RSA_CIPHER_METHODS = {  # FULL_CIPHER_NAME: (module, dict_params)
    "RSA/ECB/PKCS1Padding": (PKCS1_v1_5, {'randfunc': None}),
    "RSA/ECB/OAEPWithSHA-1AndMGF1Padding": (PKCS1_OAEP, {'hashAlgo': SHA1, 'mgfunc': lambda x, y: pss.MGF1(x, y, SHA1)}),
    "RSA/ECB/OAEPWithSHA-256AndMGF1Padding": (PKCS1_OAEP, {'hashAlgo': SHA256, 'mgfunc': lambda x, y: pss.MGF1(x, y, SHA1)}),
}


class RsaEncryptor:
    _dict_cache_cipher = {}
    _dict_cache_decipher = {}

    # ref: https://cryptobook.nakov.com/asymmetric-key-ciphers/rsa-encrypt-decrypt-examples
    def __init__(self, key_private: str = None, key_public: str = None, cipher_name: str = 'RSA/ECB/PKCS1Padding'):
        self.key_private, self.key_public = self.load_keys_from_string(key_private=key_private, key_public=key_public)
        assert self._get_cipher_module(cipher_name) is not None, 'Invalid cipher_name!'
        self.cipher_name = cipher_name

    @staticmethod
    def _get_cipher_module(full_cipher_name: str = None) -> Optional[Tuple]:
        try:
            return _RSA_CIPHER_METHODS[full_cipher_name]
        except KeyError:
            raise ValueError('Unsupported full cipher name, supported ones: %s.' % ','.join(sorted(_RSA_CIPHER_METHODS)))

    @staticmethod
    def generate_key_pair(size: int = 1024) -> Tuple[str, str]:
        key_pair = RSA.generate(size)
        key_private, key_public = key_pair.exportKey(), key_pair.publickey().exportKey()
        return key_private.decode('ascii'), key_public.decode('ascii')

    @lru_cache
    def get_cipher(self, key_public: str = None, cipher_name='RSA/ECB/PKCS1Padding') -> t_cipher_module:
        if key_public is None:
            key_pub = self.key_public
        else:
            _, key_pub = self.load_keys_from_string(key_public=key_public)
        if key_pub is None:
            raise RuntimeError('Public Key not set!')
        cache_key = key_pub.export_key(format='OpenSSH')
        if cache_key not in RsaEncryptor._dict_cache_cipher:
            pkcs_module, dict_param = self._get_cipher_module(cipher_name)
            RsaEncryptor._dict_cache_cipher[cache_key] = pkcs_module.new(key_pub, **dict_param)
        # debug: print('->PUB',  cache_key, len(RsaEncryptor._dict_cache_cipher))
        return RsaEncryptor._dict_cache_cipher[cache_key]

    @lru_cache
    def get_decipher(self, key_private: str = None, cipher_name='RSA/ECB/PKCS1Padding') -> t_cipher_module:
        if key_private is None:
            key_pri = self.key_private
        else:
            key_pri, _ = self.load_keys_from_string(key_private=key_private)
        if key_pri is None:
            raise RuntimeError('Private Key not set!')
        cache_key = key_pri.export_key(format='OpenSSH')
        if cache_key not in RsaEncryptor._dict_cache_decipher:
            pkcs_module, dict_param = self._get_cipher_module(cipher_name)
            RsaEncryptor._dict_cache_decipher[cache_key] = pkcs_module.new(key_pri, **dict_param)
        # debug: print('->PRI', cache_key, len(RsaEncryptor._dict_cache_decipher))
        return RsaEncryptor._dict_cache_decipher[cache_key]

    @staticmethod
    def load_keys_from_binary(key_private: bytes = None, key_public: bytes = None) -> Tuple[Optional[RSA.RsaKey], Optional[RSA.RsaKey]]:
        _key_private, _key_public = None, None

        if key_private is not None:
            try:
                _key_private = RSA.import_key(key_private)
            except ValueError as e:
                raise ValueError('RSA pri key format error: [%s]' % key_private)

        if key_public is not None:
            try:
                _key_public = RSA.import_key(key_public)
            except ValueError as e:
                raise ValueError('RSA pub key format error: [%s]' % key_public)

        return _key_private, _key_public

    @staticmethod
    def load_keys_from_string(key_private: str = None, key_public: str = None) -> Tuple[Optional[RSA.RsaKey], Optional[RSA.RsaKey]]:
        _key_private, _key_public = None, None

        if key_private is not None:
            if not key_private.startswith('-----'):
                key_pri = f'-----BEGIN RSA PRIVATE KEY-----\n{key_private}\n-----END RSA PRIVATE KEY-----'
            else:
                key_pri = key_private
            try:
                _key_private = RSA.import_key(key_pri)
            except ValueError as e:
                raise ValueError('RSA private key format error: [%s]' % key_pri)

        if key_public is not None:
            if not key_public.startswith('-----'):
                key_pub = f'-----BEGIN PUBLIC KEY-----\n{key_public}\n-----END PUBLIC KEY-----'
            else:
                key_pub = key_public
            try:
                _key_public = RSA.import_key(key_pub)
            except ValueError as e:
                raise ValueError('RSA public key format error: [%s]' % key_pub)

        return _key_private, _key_public

    def encrypt_with_public_key(self, message: Union[str, bytes], key_public: str = None, cipher_name: str = None) -> bytes:
        data = message if isinstance(message, bytes) else message.encode('UTF-8')
        cipher = self.get_cipher(key_public=key_public, cipher_name=cipher_name or self.cipher_name)
        return cipher.encrypt(data)

    def decrypt_with_private_key(self, ciphertext: Union[str, bytes], key_private: str = None, cipher_name: str = None, **kwargs) -> bytes:
        data = ciphertext if isinstance(ciphertext, bytes) else ciphertext.encode('ascii')
        decipher = self.get_decipher(key_private=key_private, cipher_name=cipher_name or self.cipher_name)
        return decipher.decrypt(data, **kwargs)

    @staticmethod
    def convert_bytes_to_base64(data: bytes) -> str:
        return base64.b64encode(data).decode()

    @staticmethod
    def convert_base64_to_bytes(data: str) -> bytes:
        return base64.decodebytes(data.encode('ascii'))


def main():
    key_pairs = [  # private_key, public_key
        RsaEncryptor.generate_key_pair(),
    ]

    for str_src in ('aloha!', '😄', '{"x": 1}'):
        for i, (key_pri, key_pub) in enumerate(key_pairs):
            rsa_enc = RsaEncryptor()
            x_enc = rsa_enc.encrypt_with_public_key(str_src, key_public=key_pub)
            x_txt = rsa_enc.convert_bytes_to_base64(x_enc)

            rsa_dec = RsaEncryptor()
            y_bin = rsa_dec.convert_base64_to_bytes(x_txt)
            y_dec = rsa_dec.decrypt_with_private_key(y_bin, key_private=key_pri)
            y_txt = y_dec.decode('UTF-8')

            print('[test {i_case} success = {status}] {src} -> {enc}'.format(
                i_case=i, status=(y_txt == str_src), src=y_txt, enc=x_txt
            ))



import base64
import binascii
from typing import Union

from Crypto.Cipher import AES
from Crypto.Random import get_random_bytes
from Crypto.Util.Padding import pad, unpad

_AES_CIPHER_METHODS = {  # FULL_CIPHER_NAME: (dict_params,)
    "AES/ECB/PKCS5Padding": {'mode': AES.MODE_ECB},
    # "AES/ECB/NoPadding": {'mode': AES.MODE_ECB},
    # "AES/CBC/PKCS5Padding": {'mode': AES.MODE_CBC, 'iv': b'0000000000000000'},
    # "AES/CBC/NoPadding": {'mode': AES.MODE_CBC, 'iv': b'0000000000000000'},
}


def _generate_key(key_size: int, method='const') -> bytes:
    if method == 'const':
        return b'0' * key_size  # b'b6046801716aec00'
    elif method == 'random':
        return get_random_bytes(key_size)
    raise ValueError('Invalid AES key generate method: [%s]' % method)


class AesEncryptor:
    def __init__(self, key: Union[str, bytes] = None, key_size: int = 16, cipher_name: str = 'AES/ECB/PKCS5Padding'):
        _key = key
        if key is None:
            _key = _generate_key(key_size)
        elif isinstance(key, str):
            _key = key.encode()

        if len(_key) not in (16, 24, 32,):
            raise ValueError("Invalid key size/length [%s] for AesEncryptor!" % len(_key))

        self.key_aes, self.block_size = _key, AES.block_size
        # https://pycryptodome.readthedocs.io/en/latest/src/util/util.html
        self.cipher_name = cipher_name

    def encrypt(self, text: str, output_format='hex') -> Union[str, bytes]:
        padded = pad(text.encode(), block_size=self.block_size)

        dict_params = _AES_CIPHER_METHODS.get(self.cipher_name)
        cipher = AES.new(key=self.key_aes, **dict_params)
        bytes_crypt = cipher.encrypt(padded)

        if output_format == 'hex':
            crypt = binascii.b2a_hex(bytes_crypt).decode()
        elif output_format == 'base64':
            crypt = base64.b64encode(bytes_crypt).decode()
        elif output_format in ('bytes', 'bin'):
            crypt = bytes_crypt
        else:
            raise ValueError('Unknown output_type [%s]' % output_format)
        return crypt

    def decrypt(self, text: Union[str, bytes], input_format: str = 'hex') -> Union[str, bytes]:
        text += (len(text) % 4) * '='
        if input_format == 'hex':
            crypt = binascii.a2b_hex(text)
        elif input_format == 'base64':
            crypt = base64.b64decode(text)
        elif input_format in ('bytes', 'bin'):
            crypt = text
        else:
            raise ValueError('Unknown output_type [%s]' % input_format)
        dict_params = _AES_CIPHER_METHODS.get(self.cipher_name)
        cipher = AES.new(key=self.key_aes, **dict_params)
        data = cipher.decrypt(crypt)
        data = unpad(data, block_size=self.block_size)
        return data.decode('UTF-8')


def main():
    a = AesEncryptor()
    src = 'hello~'
    enc = a.encrypt(src, output_format='base64')
    dec = a.decrypt(enc, input_format='base64')
    # print(enc, dec, src == dec)



import requests
from requests.adapters import HTTPAdapter, Retry

from ...logger import LOG


class APICaller:
    retry_method_whitelist = frozenset(['GET', 'POST'])
    retry_status_forcelist = frozenset({413, 429, 503, 502, 504})

    @classmethod
    def get_request_session(cls, total_retries: int = 3, *args, **kwargs) -> requests.Session:
        session = requests.Session()
        # https://urllib3.readthedocs.io/en/latest/reference/urllib3.util.html#urllib3.util.Retry.DEFAULT_ALLOWED_METHODS
        retries = Retry(
            total=total_retries, backoff_factor=0.1, method_whitelist=cls.retry_method_whitelist, status_forcelist=cls.retry_status_forcelist
        )
        for prefix in ('http://', 'https://'):
            session.mount(prefix, HTTPAdapter(max_retries=retries))
        return session

    @staticmethod
    def wrap_request_data(data):
        assert isinstance(data, dict), "Data object must be a dict!"
        return data

    def call(self, api_url, timeout=5, **kwargs):
        """Trigger API call

        :param api_url: do NOT start with slash (/)
        :param timeout: requests timeout in seconds
        :param kwargs: request data
        :return:
        """
        payload = APICaller.wrap_request_data(data=kwargs)
        LOG.debug('Calling API: %s' % api_url)
        session = self.get_request_session()
        resp = session.post(
            api_url, json=payload, timeout=timeout, headers={'Content-Type': 'application/json'}
        )

        try:
            ret = resp.json()
        except Exception as e:
            LOG.error(str(e))
            raise RuntimeError(resp.text)

        try:
            return ret['data']
        except KeyError:
            raise RuntimeError(resp.text)
